# -*- coding: utf-8 -*-
##############################################################################
#
#    wms_scanner module for OpenERP, Module for manage barcode reader
#    Copyright (C) 2010-2011 SYLEAM (<http://syleam.fr/>)
#              Christophe Chauvet <christophe.chauvet@syleam.fr>
#
#    This file is a part of wms_scanner
#
#    wms_scanner is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    wms_scanner is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from osv import osv
from osv import fields
from tools.translate import _
import uuid
import netsvc
from tools import config
from threading import Semaphore

logger = netsvc.Logger()


class ScannerScenario(osv.osv):
    _name = 'scanner.scenario'
    _description = 'Scenario for scanner'

    _columns = {
        'name': fields.char('Name', size=64, help='Appear on barcode reader screen'),
        'sequence': fields.integer('Sequence'),
        'active': fields.boolean('Active', help='if check, this object is always available'),
        'model_id': fields.many2one('ir.model', 'Model', required=True),
        'step_ids': fields.one2many('scanner.scenario.step', 'scenario_id', 'Scenario', ),
        'warehouse_ids': fields.many2many('stock.warehouse', 'scanner_scenario_warehouse_rel', 'scenario_id', 'warehouse_id', 'Warehouses'),
        'title': fields.char('Title', size=256, help='Title for this scenario'),
        'note': fields.text('Note', help='Store different note, date and title for modification, etc...'),
        'resid': fields.char('Reference id', size=64, readonly=True, help="Use by import/export Scenario"),
        'shared_custom': fields.boolean('Shared Chustom', help="Allows to shared the custom values with a shared scanner in the same warehouse"),
    }

    _order = 'sequence'

    _defaults = {
        'title': lambda *a: False,
    }

    _sql_constraints = [
        ('resid_uniq', 'unique (resid)', 'The resid of the Step must be unique !'),
    ]

    ## dict to save current semarophe {scenario: {warehouse: {refdoc: instance of semaphore}}}
    _semaphore = {}

    def create(self, cr, uid, values, context=None):
        """
        if the resid not in value, we will create with uuid1
        """
        if not values.get('resid'):
            values['resid'] = uuid.uuid1()
        return super(ScannerScenario, self).create(cr, uid, values, context=context)

    def _semaphore_acquire(self, cr, uid, id, warehouse_id, ref_doc, context=None):
        """
        make an acquire on a semaphore to take one jeton
        the semaphore is use like a mutex one on one
        """
        sem = self._semaphore.get(id, {}).get(warehouse_id, {}).get(ref_doc, None)
        if not sem:
            sem = Semaphore()
            if not self._semaphore.get(id):
                self._semaphore[id] = {}
            if not self._semaphore[id].get(warehouse_id):
                self._semaphore[id][warehouse_id] = {}
            self._semaphore[id][warehouse_id][ref_doc] = sem
        sem.acquire()

    def _semaphore_release(self, cr, uid, id, warehouse_id, ref_doc, context=None):
        """
        make a release on a semaphore to take one jeton
        the semaphore is use like a mutex one on one
        """
        self._semaphore[id][warehouse_id][ref_doc].release()

ScannerScenario()


class ScannerScenarioCustom(osv.osv):
    _name = 'scanner.scenario.custom'
    _description = 'Tempori value for scenario'

    _columns = {
        ### Link data to scenario
        'scenario_id': fields.many2one('scanner.scenario', 'Scenarii', required=True, help="Values use for this scenario"),
        'scanner_id': fields.many2one('scanner.hardware', 'Scanner', required=True, help="Values use for this scanner"),
        'model': fields.char('Model', size=255, required=True, help="Model use for data"),
        'res_id': fields.integer('Values id', required="True", help="Id of the model source"),
        ### Fields tmp
        'char_val1': fields.char('Char Val1', size=255, ),
        'char_val2': fields.char('Char Val2', size=255, ),
        'char_val3': fields.char('Char Val3', size=255, ),
        'char_val4': fields.char('Char Val4', size=255, ),
        'char_val5': fields.char('Char Val5', size=255, ),
        'int_val1': fields.integer('Int Val1',),
        'int_val2': fields.integer('Int Val2'),
        'int_val3': fields.integer('Int Val3'),
        'int_val4': fields.integer('Int Val4'),
        'int_val5': fields.integer('Int Val5'),
        'float_val1': fields.float('Float Val1', digits=(16, int(config['price_accuracy']))),
        'float_val2': fields.float('Float Val2', digits=(16, int(config['price_accuracy']))),
        'float_val3': fields.float('Float Val3', digits=(16, int(config['price_accuracy']))),
        'float_val4': fields.float('Float Val4', digits=(16, int(config['price_accuracy']))),
        'float_val5': fields.float('Float Val5', digits=(16, int(config['price_accuracy']))),
        'text_val': fields.text('Text Val', ),
    }

    _defaults = {
         'char_val1': lambda *a: '',
         'char_val2': lambda *a: '',
         'char_val3': lambda *a: '',
         'char_val4': lambda *a: '',
         'char_val5': lambda *a: '',
         'int_val1': lambda *a: 0,
         'int_val2': lambda *a: 0,
         'int_val3': lambda *a: 0,
         'int_val4': lambda *a: 0,
         'int_val5': lambda *a: 0,
         'float_val1': lambda *a: 0.0,
         'float_val2': lambda *a: 0.0,
         'float_val3': lambda *a: 0.0,
         'float_val4': lambda *a: 0.0,
         'float_val5': lambda *a: 0.0,
         'text_val': lambda *a: '',
    }

    def _get_domain(self, cr, uid, scenario, scanner, context=None):
        """
        create a domain to found custom values.
        use the fields shared_custom of scenario and scanner
        """
        if scenario.shared_custom == True:
            return [('scenario_id', '=', scenario.id), ('scanner_id.ref_doc', '=', scanner.ref_doc), ('scanner_id.warehouse_id', '=', scanner.warehouse_id.id)]
        return [('scenario_id', '=', scenario.id), ('scanner_id', '=', scanner.id)]

    def _get_values(self, cr, uid, scenario, scanner, model='', res_id=None, domain=None, context=None):
        """
        return read customs line
        @param domain : list of tuple for search method
        """
        if domain is None:
            domain = self._get_domain(cr, uid, scenario, scanner, context=context)
        else:
            domain.extend(self._get_domain(cr, uid, scenario, scanner, context=context))
        if model:
            domain.append(('model', '=', model))
        if res_id:
            domain.append(('res_id', '=', res_id))
        ids = self.search(cr, uid, domain, context=context)
        if ids:
            return self.read(cr, uid, ids, [], context=context)
        return []

    def _set_values(self, cr, uid, values, context=None):
        """
        values is a dict read. get id in values and delete of dict unchanging field
        """
        vals = values.copy()
        ids = [vals.get('id')]
        try:
            del vals['id']
            del vals['scenario_id']
            del vals['scanner_id']
            del vals['model']
            del vals['res_id']
        except:
            pass
        return self.write(cr, uid, ids, vals, context=context)

    def _remove_values(self, cr, uid, scenario, scanner, context=None):
        """
        Unlink all the line link we one scenarii and one scanner
        """
        domain = []
        scanner_ids = []
        if scenario.shared_custom == True:
            scanner_ids = self.pool.get('scanner.hardware').search(cr, uid, [('scenario_id', '=', scenario.id), ('warehouse_id', '=', scanner.warehouse_id.id), ('ref_doc', '=', scanner.ref_doc), ('id', '!=', scanner.id)], context=context)
        domain = [('scenario_id', '=', scenario.id), ('scanner_id', '=', scanner.id)]
        ids = self.search(cr, uid, domain, context=context)
        if scanner_ids:
            return self.write(cr, uid, ids, {'scanner_id': scanner_ids[0]}, context=context)
        else:
            return self.unlink(cr, uid, ids, context=context)


ScannerScenarioCustom()


class ScannerScenarioStep(osv.osv):
    _name = 'scanner.scenario.step'
    _description = 'Step for scenario'

    _columns = {
        'name': fields.char('Name', size=64, help='Name of the step'),
        'scenario_id': fields.many2one('scanner.scenario', 'Scenario', required=True),
        'step_start': fields.boolean('Step start', help='Check this if this is the first step of scenario'),
        'step_stop': fields.boolean('Step stop', help='check this if this is the last step of scenario'),
        'out_trans_ids': fields.one2many('scanner.scenario.transition', 'from_id', 'Outgoing transitions'),
        'in_trans_ids': fields.one2many('scanner.scenario.transition', 'to_id', 'Incoming transitions'),
        'python_code': fields.text('Python code', ),
        'resid': fields.char('Reference id', size=64, readonly=True, help="Use by import/export Scenario/Step"),
    }

    _defaults = {
        'step_start': lambda *a: False,
        'step_stop': lambda *a: False,
        'python_code': lambda *a: '# Use message to retrieve the data transmit by the scanner\n# res containt the result',
    }

    _sql_constraints = [
        ('resid_uniq', 'unique (resid)', 'The resid of the Step must be unique !'),
    ]

    def create(self, cr, uid, values, context=None):
        """
        if the resid not in value, we will create with uuid1
        """
        if not values.get('resid'):
            values['resid'] = uuid.uuid1()
        return super(ScannerScenarioStep, self).create(cr, uid, values, context=context)

ScannerScenarioStep()


class ScannerScenarioTransition(osv.osv):
    _name = 'scanner.scenario.transition'
    _description = 'Transition for scenario'

    _columns = {
        'name': fields.char('Name', size=64, required=True),
        'from_id': fields.many2one('scanner.scenario.step', 'From', required=True),
        'to_id': fields.many2one('scanner.scenario.step', 'To', required=True),
        'condition': fields.char('Condition', size=256, required=True),
        'trans_type': fields.selection([('scan', 'Scan'), ('key', 'Keyboard')], 'Transition Type'),
        'tracer': fields.char('Tracer', size=12, help='Use to determine, on which transition we arrive to this step'),
        'resid': fields.char('Reference id', size=64, readonly=True, help="Use by import/export Scenario/Transition"),
    }

    _defaults = {
        'condition': lambda *a: 'True',
        'trans_type': lambda *a: 'scan',
        'tracer': lambda *a: False,
    }

    _sql_constraints = [
        ('resid_uniq', 'unique (resid)', 'The resid of the Transition must be unique !'),
    ]

    def create(self, cr, uid, values, context=None):
        """
        if the resid not in value, we will create with uuid1
        """
        if not values.get('resid'):
            values['resid'] = uuid.uuid1()
        return super(ScannerScenarioTransition, self).create(cr, uid, values, context=context)

ScannerScenarioTransition()


class ScannerHardware(osv.osv):
    _name = 'scanner.hardware'
    _description = 'Scanner Hardware'

    _columns = {
        'name': fields.char('Name', size=128, required=True),
        'code': fields.char('Code', size=12, required=True),
        'warehouse_id': fields.many2one('stock.warehouse', 'Warehouse', required=True),
        'scenario_id': fields.many2one('scanner.scenario', 'Scenario'),
        'step_id': fields.many2one('scanner.scenario.step', 'Current Step'),
        'ref_doc': fields.integer('Reference'),
        'tmp_val1': fields.char('Temp value 1', size=128, ),
        'tmp_val2': fields.char('Temp value 2', size=128, ),
        'tmp_val3': fields.char('Temp value 3', size=128, ),
        'tmp_val4': fields.char('Temp value 4', size=128, ),
        'tmp_val5': fields.char('Temp value 5', size=128, ),
    }

    _defaults = {
        'scenario_id': lambda *a: False,
        'step_id': lambda *a: False,
        'ref_doc': lambda *a: False,
    }

    def scanner_check(self, cr, uid, numterm):
        """
        Check the step for this scanner
        """
        context = self.pool.get('res.users').context_get(cr, uid)
        term_id = self.search(cr, uid, [('code', '=', numterm)])
        if not term_id:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'Terminal %s not exists!' % numterm)
            raise osv.except_osv(_('Error'), _('This terminal not exists!'))
        term = self.browse(cr, uid, term_id[0])
        if term.scenario_id:
            return True
        return False

    def scanner_call(self, cr, uid, numterm=None, action=None, message='', etype='K'):
        """
        This method is the only one, that called by the barcode reader,

        :param etype: Indicate K for keyboard, S for Scan, use for transition
        :type  etype: str
        :return: A tuple for the action
        :rtype: tuple
        """
        context = self.pool.get('res.users').context_get(cr, uid)
        logger.notifyChannel('wms_scanner', netsvc.LOG_DEBUG, 'SCANNER_CALL')
        if numterm is None:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'scanner_call required numterm argument')
            raise osv.except_osv(_('Error'), _('Terminal number is required'))

        if action is None:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'scanner_call required action argument')
            raise osv.except_osv(_('Error'), _('Action is required'))

        term_id = self.search(cr, uid, [('code', '=', numterm)])
        if not len(term_id):
            raise osv.except_osv(_('Error'), _('This terminal is not declare!'))

        scenario = self.pool.get('scanner.scenario')
        term = self.browse(cr, uid, term_id[0])

        # we affect this terminal to the related action
        # search the schnario and the first step
        if action == 'action' and not term.scenario_id:
            if etype == 'scan':
                # there are no scenario found so no step.
                # or after we will make a browse on step with id = False and
                # it is not possible to use empty browse
                return ('U', ['Forbiden', 'action'], 0)
            logger.notifyChannel('wms_scanner', netsvc.LOG_INFO, '[%s] Action : %s (pas de scenario en cours)' % (numterm, message))
            r = self._scenario_save(cr, uid, term.id, message, etype)
            return r
        elif action == 'action' and term.scenario_id:
            r = self._scenario_save(cr, uid, term.id, message, etype, term.scenario_id.id, term.step_id.id, term.ref_doc)
            return r

        # check id terminal is not affected to a scenario and step
        # If not send the menu
        if not term.scenario_id:
            logger.notifyChannel('wms_scanner', netsvc.LOG_INFO, '[%s] Pas de scenario' % numterm)
            return self._scenario_list(cr, uid, term.warehouse_id.id)

        return ('R', ['Unknown', 'action'], 0)

    def empty_scanner_values(self, cr, uid, ids, context=None):
        """
        this method empty all tmp values, scenario, step and, ref_doc
        because if we want reset term when error we must use sql querry, it is bad in production
        """
        custom_obj = self.pool.get('scanner.scenario.custom')
        for scanner in self.browse(cr, uid, ids, context=context):
            if scanner.scenario_id and scanner.scenario_id.id:
                custom_obj._remove_values(cr, uid, scanner.scenario_id, scanner, context=context)
        self.write(cr, uid, ids, {'scenario_id': False, 'step_id': False, 'ref_doc': 0, 'tmp_val1': '', 'tmp_val2': '', 'tmp_val3': '', 'tmp_val4': '', 'tmp_val5': ''}, context=context)
        return True

    def scanner_end(self, cr, uid, numterm=None, context=None):
        """
        End the end barcode is read, we execute this step
        """
        if context is None:
            context = self.pool.get('res.users').context_get(cr, uid)

        term_id = self.search(cr, uid, [('code', '=', numterm)])
        if not term_id:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'Terminal %s not exists!' % numterm)
            raise osv.except_osv(_('Error'), _('This terminal not exists!'))

        logger.notifyChannel('wms_scanner', netsvc.LOG_INFO, 'End scenario request for %s terminal' % numterm)
        self.empty_scanner_values(cr, uid, term_id, context=context)

        return ('F', ['This scenario', 'is finished'], '')

    def _memorize(self, cr, uid, term_id, scenario_id, step_id, object=None, context=None):
        """
        After affect a scenario to a scanner, we must memorize it
        If object is specify, save it as well (ex: res.partner,12)
        """
        if context is None:
            context = self.pool.get('res.users').context_get(cr, uid)

        sca = self.pool.get('scanner.hardware')
        sca_args = {'scenario_id': scenario_id, 'step_id': step_id}
        if object and isinstance(object, int):
            sca_args['ref_doc'] = object
        sca.write(cr, uid, [term_id], sca_args)

    def _scenario_save(self, cr, uid, term_id, message, etype, scenario_id=None, step_id=None, cur_obj='', context=None):
        """
        Save the scenario on this terminal and execute the current step
        Return the action to the terminal
        """
        if context is None:
            context = self.pool.get('res.users').context_get(cr, uid)

        if etype == 'K':
            etype = 'key'
        else:
            etype = 'scan'

        scenario = self.pool.get('scanner.scenario')
        step = self.pool.get('scanner.scenario.step')
        term = self.pool.get('scanner.hardware')
        old_step = False
        # No scenario, ok we search it
        tracer = False
        if not scenario_id:
            # check with the warehouse because we can t use a scenario of other warehouse
            term_warehouse_id = term.read(cr, uid, term_id, ['warehouse_id'], context=context).get('warehouse_id', False)
            if term_warehouse_id:
                # take only the id
                term_warehouse_id = term_warehouse_id[0]
            scenario_ids = scenario.search(cr, uid, [('name', '=', message), ('warehouse_ids', 'in', [term_warehouse_id])])
            if scenario_ids:
                scenario_id = scenario_ids[0]
                step_ids = step.search(cr, uid, [('scenario_id', '=', scenario_id), ('step_start', '=', True)])
                if not step_ids:
                    return ('U', ['please contact', 'your', 'administrator', 'A001'], 0)
                step_id = step_ids[0]
            else:
                # there are no scenario found so no step.
                # or after we will make a browse on step with id = False and
                # it is not possible to use empty browse
                return ('U', ['Forbiden', 'action'], 0)
        else:
            # we search the next step for the current scenario
            old_step = step_id
            trans = self.pool.get('scanner.scenario.transition')
            trs_ids = trans.search(cr, uid, [('from_id', '=', step_id)])
            for transition in trans.browse(cr, uid, trs_ids, context=context):
                step_id = False
                tracer = ''
                cxt = {
                    'context': context,
                    'model': self.pool.get(transition.from_id.scenario_id.model_id.model),
                    'cr': cr,
                    'pool': self.pool,
                    'uid': uid,
                    'message': message,
                    't': term.browse(cr, uid, term_id, context=context),
                }
                expr = eval(str(transition.condition), cxt)

                if not expr:
                    continue
                step_id = transition.to_id.id
                tracer = transition.tracer
                break

            if not step_id:
                return ('U', ['Please contact', 'your', 'administrator'], 0)

        self._memorize(cr, uid, term_id, scenario_id, step_id, cur_obj, context)
        # MUTEXT
        t = term.browse(cr, uid, term_id, context=context)
        scenario._semaphore_acquire(cr, uid, t.scenario_id.id, t.warehouse_id.id, t.ref_doc, context=context)
        # execute the step and return the result
        stp = step.browse(cr, uid, step_id, context=context)
        logger.notifyChannel('wms_scanner', netsvc.LOG_INFO, 'Model: %s' % stp.scenario_id.model_id.model)
        ld = {
            'cr': cr,
            'uid': uid,
            'pool': self.pool,
            'model': self.pool.get(stp.scenario_id.model_id.model),
            'custom': self.pool.get('scanner.scenario.custom'),
            'term': term,
            'context': context,
            'm': message,
            't': t,
            'tracer': tracer,
            'wkf': netsvc.LocalService('workflow'),
            'scenario': self.pool.get('scanner.scenario').browse(cr, uid, scenario_id, context=context),
        }
        try:
            exec stp.python_code in ld
            if stp.step_stop:
                self.empty_scanner_values(cr, uid, [term_id], context=context)
        except osv.except_osv, e:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'OSV Exception: %s' % str(e))
            ld.update({'act': 'U', 'res': ['Please contact', 'your', 'administrator'], 'val': 0})
        except Exception, e:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'Exception: %s' % str(e))
            ld.update({'act': 'U', 'res': ['Please contact', 'your', 'administrator'], 'val': 0})
        finally:
            scenario._semaphore_release(cr, uid, t.scenario_id.id, t.warehouse_id.id, t.ref_doc, context=context)
        return (ld.get('act', 'M'), ld.get('res', ['nothing']), ld.get('val', 0))

    def _scenario_list(self, cr, uid, warehouse_id):
        """
        Retrieve the scenario list for this warehouse
        """
        # List scenario for this material
        context = self.pool.get('res.users').context_get(cr, uid)
        scenario = self.pool.get('scanner.scenario')
        scenario_ids = scenario.search(cr, uid, [('warehouse_ids', 'in', [warehouse_id])])
        scene = scenario.read(cr, uid, scenario_ids, ['name'], context=context)

        ret = []
        for s in scene:
            ret.append(s['name'])
        return ('M', ret, 0)

ScannerHardware()

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:
